%

%
% Copyright (c) 2015,2017 LAAS/CNRS
% All rights reserved.
%
% Redistribution  and  use  in  source  and binary  forms,  with  or  without
% modification, are permitted provided that the following conditions are met:
%
%   1. Redistributions of  source  code must retain the  above copyright
%      notice and this list of conditions.
%   2. Redistributions in binary form must reproduce the above copyright
%      notice and  this list of  conditions in the  documentation and/or
%      other materials provided with the distribution.
%
% THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
% WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
% MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
% ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
% WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
% ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
% IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
%
%                                           Anthony Mallet on Fri Jun 26 2015
%
function tk3_block(block)
  setup(block);
end

function setup(block)
  tk3_mask(block.BlockHandle, 0);
  user = get_param(block.BlockHandle, 'UserData');

  % Register number of ports
  block.NumInputPorts  = user.inports;
  block.NumOutputPorts = user.outports;
  block.SetPreCompInpPortInfoToDynamic;
  block.SetPreCompOutPortInfoToDynamic;

  % default input/output port properties
  for i = 1:user.inports
    block.InputPort(i).Dimensions = 1;
    block.InputPort(i).DatatypeID = 0; % double
    block.InputPort(i).Complexity = 'Real';
    block.InputPort(i).SampleTime = [-1 0];
  end
  for i = 1:user.outports
    block.OutputPort(i).DimensionsMode = 'Fixed';
    block.OutputPort(i).Dimensions = 1;
    block.OutputPort(i).DatatypeID = 0; % double
    block.OutputPort(i).Complexity = 'Real';
    block.OutputPort(i).SamplingMode = 'Sample';
  end

  % specific input/output ports
  if user.inidx.start >= 0
    block.InputPort(user.inidx.start).DatatypeID = 8; % boolean
  end
  if user.outidx.imu >= 0
    f = str2num(user.param.imufreq);
    block.OutputPort(user.outidx.imu).Dimensions = 6;
    block.OutputPort(user.outidx.imu).SampleTime = [1/f 0];
  end
  if user.outidx.imuts >= 0
    f = str2num(user.param.imufreq);
    block.OutputPort(user.outidx.imuts).SampleTime = [1/f 0];
  end
  if user.outidx.vel >= 0
    f = str2num(user.param.motorfreq);
    block.OutputPort(user.outidx.vel).SampleTime = [1/f 0];
  end
  if user.outidx.pwm >= 0
    f = str2num(user.param.motorfreq);
    block.OutputPort(user.outidx.pwm).SampleTime = [1/f 0];
  end
  if user.outidx.cur >= 0
    f = str2num(user.param.motorfreq);
    block.OutputPort(user.outidx.cur).SampleTime = [1/f 0];
  end
  if user.outidx.mts >= 0
    f = str2num(user.param.motorfreq);
    block.OutputPort(user.outidx.mts).SampleTime = [1/f 0];
  end
  if user.outidx.bat >= 0
    f = str2num(user.param.batteryfreq);
    block.OutputPort(user.outidx.bat).SampleTime = [1/f 0];
  end
  if user.outidx.batts >= 0
    f = str2num(user.param.batteryfreq);
    block.OutputPort(user.outidx.batts).SampleTime = [1/f 0];
  end
  if user.outidx.mid >= 0
    f = str2num(user.param.motorfreq);
    block.OutputPort(user.outidx.mid).SampleTime = [1/f 0];
  else
    if user.inidx.cmd >= 0
      block.InputPort(user.inidx.cmd).Dimensions = str2num(user.param.nmotor);
    end
    if user.inidx.thr >= 0
      block.InputPort(user.inidx.thr).Dimensions = str2num(user.param.nmotor);
    end
    if user.outidx.vel >= 0
      block.OutputPort(user.outidx.vel).Dimensions = str2num(user.param.nmotor);
    end
    if user.outidx.pwm >= 0
      block.OutputPort(user.outidx.pwm).Dimensions = str2num(user.param.nmotor);
    end
    if user.outidx.cur >= 0
      block.OutputPort(user.outidx.cur).Dimensions = str2num(user.param.nmotor);
    end
    if user.outidx.mts >= 0
      block.OutputPort(user.outidx.mts).Dimensions = str2num(user.param.nmotor);
    end
  end

  % register methods
  block.SimStateCompliance = 'HasNoSimState';

  block.RegBlockMethod('SetInputPortSamplingMode', @setinportsamplingmode);
  block.RegBlockMethod('SetInputPortSampleTime', @setinputportsampletime);
  block.RegBlockMethod('SetOutputPortSampleTime', @setoutputportsampletime);
  block.RegBlockMethod('Start', @start);
  if user.outidx.imu >= 0 || user.outidx.vel >= 0 || user.outidx.bat >= 0
    block.RegBlockMethod('Outputs', @outputs);
  end
  if user.inidx.start >= 0 || user.inidx.cmd >= 0 || user.inidx.thr >= 0
    block.RegBlockMethod('Update', @update);
  end
  block.RegBlockMethod('Terminate', @terminate);
end

function setinportsamplingmode(block, idx, fd)
  block.InputPort(idx).SamplingMode = fd;
end

function setinputportsampletime(block, idx, time)
  block.InputPort(idx).SampleTime = time;
end
function setoutputportsampletime(block, idx, time)
  % outport frequency should not be changed
end

%%
%% Start:
%%   Functionality    : Called once at start of model execution. If you
%%                      have states that should be initialized once, this
%%                      is the place to do it.
%%   Required         : No
%%   C-MEX counterpart: mdlStart
%%
function start(block)
  user = get_param(block.BlockHandle, 'UserData');
  switch user.param.device
    case 'Flight controller'
      dev = 'mkfl';
    case 'Brushless controller'
      dev = 'mkbl';
  end
  user.client = tk3.client(dev, ...
                           user.param.serial, str2num(user.param.speed));

  if user.outidx.imu >= 0
    user.client.get_imu(str2num(user.param.imufreq));
  end
  if user.outidx.vel >= 0
    user.client.get_motor(str2num(user.param.motorfreq));
  end
  if user.outidx.bat >= 0
    user.client.get_battery(str2num(user.param.batteryfreq));
  end
  user.client.stop;
  user.start = 0;

  if user.outidx.vel >= 0
     for i = 1:block.OutputPort(user.outidx.vel).Dimensions
       block.OutputPort(user.outidx.vel).Data(i) = NaN;
     end
  end
  if user.outidx.pwm >= 0
     for i = 1:block.OutputPort(user.outidx.pwm).Dimensions
       block.OutputPort(user.outidx.pwm).Data(i) = NaN;
     end
  end
  if user.outidx.cur >= 0
     for i = 1:block.OutputPort(user.outidx.cur).Dimensions
       block.OutputPort(user.outidx.cur).Data(i) = NaN;
     end
  end
  if user.outidx.mts >= 0
     for i = 1:block.OutputPort(user.outidx.mts).Dimensions
       block.OutputPort(user.outidx.mts).Data(i) = NaN;
     end
  end

  set_param(block.BlockHandle, 'UserData', user);
end

function outputs(block)
  user = get_param(block.BlockHandle, 'UserData');
  if isempty(fieldnames(user.outidx)) return; end

  data = user.client.data;

  if user.outidx.imu >= 0 && ~isempty(data.acc_x)
    block.OutputPort(user.outidx.imu).Data = ...
    [ data.acc_x, data.acc_y, data.acc_z, ...
      data.gyr_r, data.gyr_p, data.gyr_y ];
    if user.outidx.imuts >= 0
      block.OutputPort(user.outidx.imuts).Data = data.imu_ts;
    end
  end

  if user.outidx.mid >= 0
    if ~isempty(data.id)
      block.OutputPort(user.outidx.mid).Data = data.id;
      if user.outidx.vel >= 0
        block.OutputPort(user.outidx.vel).Data = data.velocity;
      end
      if user.outidx.pwm >= 0
        block.OutputPort(user.outidx.pwm).Data = data.pwm;
      end
      if user.outidx.cur >= 0
        block.OutputPort(user.outidx.cur).Data = data.current;
      end
      if user.outidx.mts >= 0
        block.OutputPort(user.outidx.mts).Data = data.m_ts;
      end
    end
  elseif ~isempty(data.motor)
    motors = min(length(data.motor), str2num(user.param.nmotor));

    if user.outidx.vel >= 0
      block.OutputPort(user.outidx.vel).Data(1:motors) = ...
        [ data.motor(1:motors).velocity ];
    end
    if user.outidx.pwm >= 0
      block.OutputPort(user.outidx.pwm).Data(1:motors) = ...
        [ data.motor(1:motors).pwm ];
    end
    if user.outidx.cur >= 0
      block.OutputPort(user.outidx.cur).Data(1:motors) = ...
        [ data.motor(1:motors).current ];
    end
    if user.outidx.mts >= 0
      block.OutputPort(user.outidx.mts).Data(1:motors) = ...
        [ data.motor(1:motors).m_ts ];
    end
  end

  if user.outidx.bat >= 0 && ~isempty(data.battery)
    block.OutputPort(user.outidx.bat).Data = data.battery;
    if user.outidx.batts >= 0
      block.OutputPort(user.outidx.batts).Data = data.battery_ts;
    end
  end
end

function update(block)
  user = get_param(block.BlockHandle, 'UserData');

  if user.inidx.start >= 0
    n = block.InputPort(user.inidx.start).Data;
    if user.start ~= n
       user.start = n;
       if user.start
          user.client.start;
       else
          user.client.stop;
       end
       set_param(block.BlockHandle, 'UserData', user);
    end
  end
  if user.inidx.cmd >= 0
    user.client.velocity(block.InputPort(user.inidx.cmd).Data);
  end
  if user.inidx.thr >= 0
    user.client.throttle(block.InputPort(user.inidx.thr).Data);
  end
end

function terminate(block)
  user = get_param(block.BlockHandle, 'UserData');
  if isfield(user, 'client')
     delete(user.client);
  end
end
